/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;
import java.util.ArrayList;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer.Viewport} within a {@link FrameBuffer}
   data structure. This renderer mutates the {@link FrameBuffer.Viewport}
   so that it is filled in with the rendered image of the scene
   represented by the {@link Scene} object.
<p>
   This implements our fourth rendering pipeline. It is essentially the
   same as the previous renderer but it adds an antialiasing option to
   the rasterizing stage, {@link RasterizeAntialias}. So there are
   still just three pipeline stages.
*/
public class Pipeline
{
   public static boolean debug = false;

   /**
      Mutate the {@link FrameBuffer}'s given {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param vp     {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   public static void render(Scene scene, FrameBuffer.Viewport vp)
   {
      // Render every Model in the Scene.
      for (Model model : scene.modelList)
      {
         if ( model.visible )
         {
            logMessage(model, "==== Render Model: " + model.name + " ====");

            check(model);

            // 0. Make a deep copy of the model.
            Model model2 = new Model(model);

            logVertexList("0. Input    ", model2);

            // 1. Apply the projection transformation.
            Projection.project(model2.vertexList, scene.camera);

            logVertexList("1. Projected", model2);

            // 2. Clip each line segment to the camera's view rectangle.
            List<LineSegment> lineSegmentList2 = new ArrayList<>();
            for (LineSegment ls : model2.lineSegmentList)
            {
               logLineSegment("2. Clipping", model2, ls);

               if ( Clip.clip(model2, ls) )
               {
                  // Keep the line segments that are visible.
                  lineSegmentList2.add(ls);

                  logLineSegment("2. Clipping (accept)", model2, ls);
               }
               else
               {
                  logLineSegment("2. Clipping (reject)", model2, ls);
               }
            }
            // Replace the model's original list of line segments
            // with the list of clipped line segments.
            model2.lineSegmentList = lineSegmentList2;

            logVertexList("2. Clipped  ", model2);
            logLineSegmentList("2. Clipped  ", model2);

            // 3. Rasterize each visible line segment into pixels.
            for (LineSegment ls : model2.lineSegmentList)
            {
               logLineSegment("3. Rasterize", model2, ls);

               RasterizeAntialias.rasterize(model2, ls, vp);
            }

            logMessage(model2, "==== End Model: " + model2.name + " ====");
         }
         else
         {
            logMessage(model, "==== Hidden model: " + model.name + " ====");
         }
      }
   }


   /**
      Determine if there are any obvious problems with the {@link Model}
      being rendered. The purpose of these checks is to make the renderer
      a bit more user friendly. If a user makes a simple mistake and tries
      to render a {@link Model} that is missing vertices or line segments,
      then the user gets a helpful error message.

      @param model  the {@link Model} being rendered
   */
   public static void check(Model model)
   {
      boolean error = false;
      if (model.vertexList.isEmpty())
      {
         System.err.println("***WARNING: This model does not have any vertices.");
         error = true;
      }
      if (model.lineSegmentList.isEmpty())
      {
         System.err.println("***WARNING: This model does not have any line segments.");
         error = true;
      }
      if (error)
      {
         System.err.println(model);
      }
   }


   /**
      Use the pipeline's and the model's debug variables to determine
      if the given message should be printeed to stderr.

      @param model    the {@link Model} being rendered
      @param message  {@link String} to output to stderr
   */
   public static void logMessage(Model model, String message)
   {
      if (Pipeline.debug || model.debug)
         System.err.println( message );
   }


   /**
      This method prints a {@link String} representation of the given
      {@link Model}'s {@link Vertex} list.

      @param stage  name for the pipeline stage
      @param model  the {@link Model} whose vertex list is to be printed
   */
   private static void logVertexList(String stage, Model model)
   {
      if (Pipeline.debug || model.debug)
      {
         int i = 0;
         for (Vertex v : model.vertexList)
         {
            System.err.printf( stage + ": index = %3d, " + v.toString(), i );
            ++i;
         }
      }
   }


   /**
      This method prints a {@link String} representation of the given
      {@link Model}'s {@link LineSegment} list.

      @param stage  name for the pipeline stage
      @param model  the {@link Model} whose line segment list is to be printed
   */
   private static void logLineSegmentList(String stage, Model model)
   {
      if (Pipeline.debug || model.debug)
      {
         for (LineSegment ls : model.lineSegmentList)
         {
            System.err.printf( stage + ": " + ls.toString() );
         }
      }
   }


   /**
      This method prints a {@link String} representation of the given
      {@link LineSegment}.

      @param stage  name for the pipeline stage
      @param model  {@link Model} that the {@link LineSegment} {@code ls} comes from
      @param ls     {@link LineSegment} whose string representation is to be printed
   */
   private static void logLineSegment(String stage, Model model, LineSegment ls)
   {
      if (Pipeline.debug || model.debug)
      {
         System.err.print( stage + ": " + ls.toString() );
         int index0 = ls.index[0];
         int index1 = ls.index[1];
         Vertex v0 = model.vertexList.get(index0);
         Vertex v1 = model.vertexList.get(index1);
         System.err.printf("   index = %3d, " + v0.toString(), index0);
         System.err.printf("   index = %3d, " + v1.toString(), index1);
      }
   }
}
